/*
 *    This file is part of wdaemon.
 *
 *    wdaemon is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    wdaemon is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with wademon; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <sys/poll.h>
#include <linux/input.h>
#include "session.h"
#include "monitored.h"
#include "input.h"
#include "uinput.h"
#include "hotplug.h"
#include "debug.h"

int input_init(struct monitored_devices *devices)
{
	return 0;
}

int input_add_device(struct monitored_devices *devices, struct monitored_device *dev)
{
	int retval;

	dprint(1, "Opening device %s\n", dev->name);

	if (dev->tries > MAX_ADD_TRIES) {
		fprintf(stderr, "Giving up on device %s, unable to open it "
			"after %i tries\n", dev->name, MAX_ADD_TRIES);
		dev->tries = 0;
		dev->pending_action = NONE;
		dev->plugged = 0;
		dev->fd = -1;
		return 1;
	}
	dev->tries++;
	/* if the device is already active, close and reinit it */
	if (dev->fd >= 0)
		close(dev->fd);

	dev->fd = open(dev->name, O_RDWR);
	if (dev->fd < 0) {
		fprintf(stderr, "Unable to open file just created "
			"%s (%s) ", dev->name, strerror(errno));
		return 1;
	}

	/* grab the device */
	retval = ioctl(dev->fd, EVIOCGRAB, 1);
	if (retval < 0) {
		fprintf(stderr, "Unable to grab device %s (%s)\n", dev->name,
			strerror(errno));
		close(dev->fd);
		dev->fd = -1;
		return 1;
	}

	dev->tries = 0;
	dev->pending_action = NONE;
	dev->plugged = 1;

	return 0;
}

int input_remove_device(struct monitored_devices *devices, struct monitored_device *dev)
{
	dev->pending_action = NONE;
	dev->plugged = 0;
	dev->tries = 0;

	dprint(1, "Closing device %s\n", dev->name);

	if (dev->fd >= 0) {
		close(dev->fd);
		dev->fd = -1;
	}

	return 0;
}

static int input_process(struct monitored_device *dev)
{
	struct input_event ev;
	int retval, flags;

	/* switch to non blocking mode */
	retval = fcntl(dev->fd, F_GETFL);
	if (retval == -1) {
		fprintf(stderr, "Unable to get file flags from %s: %s\n",
			dev->name, strerror(errno));
		return 1;
	}
	flags = retval;
	retval = fcntl(dev->fd, F_SETFL, flags | O_NONBLOCK);
	if (retval == -1) {
		fprintf(stderr, "Unable to set file as non blocking: %s (%s)\n",
			dev->name, strerror(errno));
		return 1;
	}

	/* read and process all data */
	while((retval = read(dev->fd, &ev, sizeof(ev))) == sizeof(ev))
		if (uinput_write_event(&dev->info, &ev))
			break;

	if (retval == -EBADF || retval == -ENODEV)
		return 1;

	retval = fcntl(dev->fd, F_SETFL, flags);
	if (retval == -1) {
		fprintf(stderr, "Unable to set back the file's flags: %s (%s)\n",
			dev->name, strerror(errno));
		return 1;
	}

	return 0;
}

int input_poll(struct session *session)
{
	struct monitored_devices *devices = session->devices;
	struct monitored_device *dev;
	int retval, i, j;

	dnotify_do_pending_work(session);

	memset(devices->ufds, 0, sizeof(devices->ufds));
	for (i = 0, j = 0; i < devices->count; i++) {
		dev = &devices->device[i];

		if (dev->pending_action == ADD)
			input_add_device(devices, dev);
		else if (dev->pending_action == REMOVE)
			input_remove_device(devices, dev);

		if (dev->plugged == 0)
			continue;
		devices->ufds[j].fd = dev->fd;
		devices->ufds[j].events = POLLIN | POLLPRI;
		j++;
	}
	if (j == 0) {
		/*
		 * no available devices, just sleep a bit here to prevent busy
		 * loop
		 */
		usleep(50);
		return 1;
	}

	/* pool for device input */
	retval = poll(devices->ufds, j, 50);
	if (retval == 0)
		return 1;

	if (retval < 0) {
		perror("poll");
		return 1;
	}

	/* find the file descriptor on list */
	for (i = 0, j = 0; i < devices->count; i++) {
		dev = &devices->device[i];

		if (dev->plugged == 0)
			continue;

		if (devices->ufds[j].revents & POLLERR)
			/* device was removed */
			input_remove_device(devices, dev);
		else if (devices->ufds[j].revents & POLLIN)
			if (input_process(dev))
				input_remove_device(devices, dev);
		j++;
	}

	return 0;
}

